﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Xml;
using Common.Win.General.Security;
using DevExpress.ExpressApp;
using DevExpress.ExpressApp.Model;
using DevExpress.Persistent.Base;
using DevExpress.Persistent.BaseImpl;
using DevExpress.Xpo;
using DevExpress.Xpo.Metadata;

namespace Common.Win.General.DashBoard.BusinessObjects {
    [ImageName("BO_DashboardDefinition")]
    [DefaultProperty(nameof(DashboardDefinition.Name))]
    [DefaultClassOptions]
    [SecurityOperations("DashboardDefinitions", "DashboardOperation")]
    [NavigationItem("Reports")]
    public class DashboardDefinition : BaseObject, IDashboardDefinition {

        bool _active;
        BindingList<ITypeWrapper> _dashboardTypes;
        int _index;

        string _name;
        string _targetObjectTypes;
        IList<TypeWrapper> _types;

        public DashboardDefinition(Session session)
            : base(session) {
        }

        [ProvidedAssociation("DashboardDefinition-Roles", RelationType.ManyToMany)]
        [Browsable(false)]
        public XPCollection<XRole> Roles {
            get {
                return GetCollection<XRole>(nameof(Roles));
            }
        }

        [Browsable(false)]
        internal IList<TypeWrapper> Types {
            get {
                return _types ?? (_types = XafTypesInfo.Instance.PersistentTypes
                                                       .Where(info => (info.IsVisible && info.IsPersistent) && (info.Type != null))
                                                       .Select(info => new TypeWrapper(Session, info.Type))
                                                       .OrderBy(info => info.Caption)
                                                       .ToList());
            }
        }

        public int Index {
            get { return _index; }
            set { SetPropertyValue(nameof(Index), ref _index, value); }
        }

        public string Name {
            get { return _name; }
            set { SetPropertyValue(nameof(Name), ref _name, value); }
        }

        public bool Active {
            get { return _active; }
            set { SetPropertyValue(nameof(Active), ref _active, value); }
        }

        [Size(SizeAttribute.Unlimited)]
        [Browsable(false)]
        public string TargetObjectTypes {
            get { return _targetObjectTypes; }
            set { SetPropertyValue(nameof(TargetObjectTypes), ref _targetObjectTypes, value); }
        }

        [Size(SizeAttribute.Unlimited)]
        [Delayed]
        [VisibleInDetailView(false)]
        public string Xml {
            get { return GetDelayedPropertyValue<String>(nameof(Xml)); }
            set { SetDelayedPropertyValue(nameof(Xml), value); }
        }

        [ModelDefault("DetailViewImageEditorFixedHeight", "32"), ModelDefault("DetailViewImageEditorFixedWidth", "32"),
         ModelDefault("ListViewImageEditorCustomHeight", "32")]
        [ImmediatePostData, ValueConverter(typeof(ImageValueConverter))]
        [Delayed(true)]
        public Image Icon {
            get { return GetDelayedPropertyValue<Image>(nameof(Icon)); }
            set { SetDelayedPropertyValue(nameof(Icon), value); }
        }

        [DataSourceProperty(nameof(Types))]
        [ImmediatePostData]
        [EditorAlias("Xpand.Dashboard.Win.PropertyEditors.ChooseFromListCollectionEditor")]
        public IList<ITypeWrapper> DashboardTypes {
            get { return GetBindingList(); }
        }

        BindingList<ITypeWrapper> GetBindingList() {
            return _dashboardTypes ?? (_dashboardTypes = new BindingList<ITypeWrapper>());
        }

        public override void AfterConstruction() {
            base.AfterConstruction();
            SubscribeToListEvents();
        }


        protected override void OnLoaded() {
            base.OnLoaded();
            ReloadDashboardItems();
        }

        void UpdateTargetObjectTypes() {
            _targetObjectTypes = "<Types>\r\n";
            foreach (TypeWrapper resource in DashboardTypes)
                _targetObjectTypes += string.Format("<Value Type=\"{0}\"/>\r\n", resource.Type.FullName);
            _targetObjectTypes += "</Types>";
        }

        void ReloadDashboardItems() {
            UnsubscribeToListEvents();
            try {
                DashboardTypes.Clear();
                if (!String.IsNullOrEmpty(_targetObjectTypes)) {
                    var xmlDocument = new XmlDocument();
                    xmlDocument.LoadXml(_targetObjectTypes);
                    if (xmlDocument.DocumentElement != null)
                        foreach (XmlNode xmlNode in xmlDocument.DocumentElement.ChildNodes)
                            DashboardTypes.Add(Types.First(type => xmlNode.Attributes != null && type.Type == XafTypesInfo.Instance.FindTypeInfo(xmlNode.Attributes["Type"].Value).Type));
                }
            } finally {
                SubscribeToListEvents();
            }
        }

        void SubscribeToListEvents() {
            ((IBindingList)DashboardTypes).ListChanged += DashboardDefinition_ListChanged;
        }

        void UnsubscribeToListEvents() {
            ((IBindingList)DashboardTypes).ListChanged -= DashboardDefinition_ListChanged;
        }

        void DashboardDefinition_ListChanged(object sender, ListChangedEventArgs e) {
            if (e.ListChangedType == ListChangedType.ItemAdded || e.ListChangedType == ListChangedType.ItemDeleted) {
                UpdateTargetObjectTypes();
                OnChanged(nameof(TargetObjectTypes));
            }
        }
    }
}
