using System;
using System.ComponentModel;
using System.Drawing;
using DevExpress.ExpressApp;
using DevExpress.ExpressApp.Model;
using DevExpress.ExpressApp.Utils;
using DevExpress.Persistent.Base;
using DevExpress.Persistent.Validation;
using DevExpress.Xpo;
using DevExpress.Xpo.Metadata;
using DevExpress.Persistent.Base.General;
using Xafari.BC;
using Xafari.BC.Xpo;
using Xafari.Base;
using Xafari.BC.HierarchicalData;

namespace Sample.Classifiers
{
	/// <summary>
	/// Classifier "Employees".
	/// </summary>
	[DefaultClassOptions]
	[System.ComponentModel.Description("Classifier 'Employees'")]
	[Custom("GroupName", "Classifiers")]
	public class Employee : ClassifierItemBase, ITreeNode, ITreeNodeImageProvider, IStatusSupport
	{
		public Employee(Session session) : base(session) { }

		private Employee _parent;

		/// <summary></summary>
		[Association(typeof (Employee))]
		[ParentProperty]
		public Employee Parent
		{
			get { return _parent; }
			set { SetPropertyValue("Parent", ref _parent, value); }
		}

		public override void AfterConstruction()
		{
			base.AfterConstruction();
			this.CurrentStatus = StatusTypes.Draft;
		}

		[RuleFromBoolProperty("IsValidParent", DefaultContexts.Save, "Cannot select yourself as parent.")]
        [NonPersistent]
        [MemberDesignTimeVisibility(false)]
	    public bool IsValidParent
	    {
	        get
	        {
                var p = Parent;
                while (p != null)
                {
                    if (p == this || p == p.Parent) return false;
                    p = p.Parent;
                }
	            return true;
	        }
	    }

		/// <summary></summary>
		[Association(typeof (Employee))]
		[ChildrenProperty]
		public XPCollection<Employee> Children
		{
			get { return GetCollection<Employee>("Children"); }
		}

		/// <summary></summary>
		[Size(60)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String Address
		{
			get { return GetPropertyValue<string>("Address"); }
			set { SetPropertyValue("Address", value); }
		}
		/// <summary> </summary>

		public DateTime BirthDate
        {
            get { return GetPropertyValue<DateTime>("BirthDate"); }
            set { SetPropertyValue("BirthDate", value); }
        }
		/// <summary></summary>
        [Size(15)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String City
        {
            get { return GetPropertyValue<String>("City"); }
            set { SetPropertyValue("City", value); }
        }
		/// <summary></summary>
        [Size(15)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String Country
        {
            get { return GetPropertyValue<String>("Country"); }
            set { SetPropertyValue("Country", value); }
        }
		/// <summary> </summary>
        [Size(4)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String Extension
        {
            get { return GetPropertyValue<String>("Extension"); }
            set { SetPropertyValue("Extension", value); }
        }
		/// <summary></summary>
        [Size(10)]
		public String FirstName
        {
            get { return GetPropertyValue<String>("FirstName"); }
            set { SetPropertyValue("FirstName", value); }
        }
		/// <summary> </summary>

        public DateTime HireDate
        {
            get { return GetPropertyValue<DateTime>("HireDate"); }
            set { SetPropertyValue("HireDate", value); }
        }
		/// <summary> </summary>
        [Size(24)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String HomePhone
        {
            get { return GetPropertyValue<String>("HomePhone"); }
            set { SetPropertyValue("HomePhone", value); }
        }
		/// <summary></summary>
        [Size(20)]
		public String LastName
        {
            get { return GetPropertyValue<String>("LastName"); }
            set { SetPropertyValue("LastName", value); }
        }
		/// <summary></summary>
        [Size(SizeAttribute.Unlimited)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String Notes
        {
            get { return GetPropertyValue<String>("Notes"); }
            set { SetPropertyValue("Notes", value); }
        }
		/// <summary></summary>

		[ValueConverter(typeof(ImageValueConverter))]
		[Delayed]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public Image Photo
		{
			get { return (Image)GetDelayedPropertyValue("Photo"); }
			set { SetDelayedPropertyValue("Photo", value); }
		}
		/// <summary> </summary>
        [Size(10)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String PostalCode
        {
            get { return GetPropertyValue<String>("PostalCode"); }
            set { SetPropertyValue("PostalCode", value); }
        }
		/// <summary></summary>
        [Size(10)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String Region
        {
            get { return GetPropertyValue<String>("Region"); }
            set { SetPropertyValue("Region", value); }
        }
		/// <summary></summary>
        [Size(30)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String Title
        {
            get { return GetPropertyValue<String>("Title"); }
            set { SetPropertyValue("Title", value); }
        }
		/// <summary></summary>
        [Size(25)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public String TitleOfCourtesy
        {
            get { return GetPropertyValue<String>("TitleOfCourtesy"); }
            set { SetPropertyValue("TitleOfCourtesy", value); }
        }


		/// <summary> </summary>
		//[Xafari.BusinessFramework.Metadata.Reflection.ParentProperty]
		[PersistentAlias("Parent")]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public Employee ReportsTo
		{
			get { return Parent; }
		}

		protected override void OnDeleting()
		{
			CheckChildrenForDelete();
			base.OnDeleting();
		}

		void CheckChildrenForDelete()
		{
			var inode = (ITreeNode)this;

			if (inode.Children == null || inode.Children.Count == 0)
				return;
            throw new UserFriendlyException(CaptionHelper.GetLocalizedText(@"Texts\Employee", "DeletingNodeWithChild"));
		}

		protected override void OnChanged(string propertyName, object oldValue, object newValue)
		{

			if (propertyName == "Parent" && oldValue != null)
			{
				var _oldParent = (Employee)oldValue;
				_oldParent.Children.Remove(this);
			}
			if (propertyName == "CurrentStatus")
			{
				this.StatusModify = new ActionInfo { ActionExecutor = SecuritySystem.CurrentUserName, Date = DateTime.Now};
				//this.LastStatusModifyDate = DateTime.Now;
				//this.LastStatusModifyUser = SecuritySystem.CurrentUserName;
			}
			base.OnChanged(propertyName, oldValue, newValue);
		}

		protected override void OnSaving()
		{
			this.Name = string.Format("{0} {1}", FirstName, LastName);
			base.OnSaving();
		}

		#region ITreeNode Members

		System.ComponentModel.IBindingList ITreeNode.Children
		{
			get { return Children; }
		}

		string ITreeNode.Name
		{
			get { return Name; }
		}

		ITreeNode ITreeNode.Parent
		{
			get { return Parent; }
		}

		#endregion

		[VisibleInDetailView(false)]
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		public Employee ThisObject
		{
			get { return this; }
		}

		Image ITreeNodeImageProvider.GetImage(out string imageName)
		{
			imageName = "BO_User";
			return ImageLoader.Instance.GetImageInfo(imageName).Image;
		}

		[ModelDefault("AllowEdit", "false")]
		public StatusTypes CurrentStatus
		{
			get { return this.GetPropertyValue<StatusTypes>("CurrentStatus"); }
			set { this.SetPropertyValue("CurrentStatus", value); }
		}

		/// <summary>
		///     ,   .
		/// </summary>
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		[Browsable(false)]
		[NonPersistent]
		[Obsolete("Use StatusModify", true)]
		public string LastStatusModifyUser
		{
			get { return this.StatusModify.ActionExecutor; }
			set { this.StatusModify.ActionExecutor = value; }
		}

		/// <summary>
		///       .
		/// </summary>
		[VisibleInListView(false)]
		[VisibleInLookupListView(false)]
		[Browsable(false)]
		[NonPersistent]
		[Obsolete("Use StatusModify", true)]
		public DateTime LastStatusModifyDate
		{
			get { return this.StatusModify.Date.HasValue ? this.StatusModify.Date.Value : DateTime.MinValue; }
			set { this.StatusModify.Date = value; }
		}

		[Browsable(false)]
		public ActionInfo StatusModify;

		ActionInfo IStatusSupport.StatusModify
		{
			get { return this.StatusModify; }
			set { this.StatusModify = value; }
		}
	}
}
